import numpy as np
import cv2

from pykinect_azure.k4abt.joint2d import Joint2d
from pykinect_azure.k4abt._k4abtTypes import K4ABT_JOINT_COUNT, K4ABT_SEGMENT_PAIRS
from pykinect_azure.k4abt._k4abtTypes import (
    k4abt_skeleton2D_t,
    k4abt_body2D_t,
    body_colors,
)
from pykinect_azure.k4a._k4atypes import K4A_CALIBRATION_TYPE_DEPTH


class Body2d:
    def __init__(self, body2d_handle):
        if body2d_handle:
            self._handle = body2d_handle
            self.id = body2d_handle.id
            self.initialize_skeleton()

    def __del__(self):
        self.destroy()

    def json(self):
        return self._handle.__iter__()

    def numpy(self):
        return np.array([joint.numpy() for joint in self.joints])

    def is_valid(self):
        return self._handle

    def handle(self):
        return self._handle

    def destroy(self):
        if self.is_valid():
            self._handle = None

    def initialize_skeleton(self):
        joints = np.ndarray((K4ABT_JOINT_COUNT,), dtype=np.object_)

        for i in range(K4ABT_JOINT_COUNT):
            joints[i] = Joint2d(self._handle.skeleton.joints2D[i], i)

        self.joints = joints

    def draw(self, image, only_segments=False):
        color = (
            int(body_colors[self.id][0]),
            int(body_colors[self.id][1]),
            int(body_colors[self.id][2]),
        )

        for segmentId in range(len(K4ABT_SEGMENT_PAIRS)):
            segment_pair = K4ABT_SEGMENT_PAIRS[segmentId]
            point1 = self.joints[segment_pair[0]].get_coordinates()
            point2 = self.joints[segment_pair[1]].get_coordinates()

            if (point1[0] == 0 and point1[1] == 0) or (
                point2[0] == 0 and point2[1] == 0
            ):
                continue
            image = cv2.line(image, point1, point2, color, 2)

        if only_segments:
            return image

        for joint in self.joints:
            image = cv2.circle(image, joint.get_coordinates(), 3, color, 3)

        return image

    @staticmethod
    def create(body_handle, calibration, bodyIdx, dest_camera):
        skeleton2d_handle = k4abt_skeleton2D_t()
        body2d_handle = k4abt_body2D_t()

        for jointID, joint in enumerate(body_handle.skeleton.joints):
            skeleton2d_handle.joints2D[jointID].position = calibration.convert_3d_to_2d(
                joint.position, K4A_CALIBRATION_TYPE_DEPTH, dest_camera
            )
            skeleton2d_handle.joints2D[
                jointID
            ].confidence_level = joint.confidence_level

        body2d_handle.skeleton = skeleton2d_handle
        body2d_handle.id = bodyIdx

        return Body2d(body2d_handle)

    def __str__(self):
        """Print the current settings and a short explanation"""
        message = f"Body Id: {self.id}\n\n"

        for joint in self.joints:
            message += str(joint)

        return message



import ctypes

from . import _k4a
from .image import Image


class Transformation:
    def __init__(self, calibration):
        self.calibration = calibration
        self._handle = _k4a.k4a_transformation_create(calibration.handle())
        self.color_resolution = Resolution(
            calibration.handle().color_camera_calibration.resolution_width,
            calibration.handle().color_camera_calibration.resolution_height,
        )
        self.depth_resolution = Resolution(
            calibration.handle().depth_camera_calibration.resolution_width,
            calibration.handle().depth_camera_calibration.resolution_height,
        )

    def __del__(self):
        self.destroy()

    def is_valid(self):
        return self._handle

    def handle(self):
        return self._handle

    def destroy(self):
        if self.is_valid():
            _k4a.k4a_transformation_destroy(self._handle)
            self._handle = None

    def depth_image_to_color_camera(self, depth_image):
        if not depth_image.is_valid():
            return Image()

        transformed_depth_image = Image.create(
            depth_image.format,
            self.color_resolution.width,
            self.color_resolution.height,
            self.color_resolution.width * 2,
        )

        _k4a.k4a_transformation_depth_image_to_color_camera(
            self._handle, depth_image.handle(), transformed_depth_image.handle()
        )

        return transformed_depth_image

    def depth_image_to_color_camera_custom(
        self,
        depth_image,
        custom_image,
        interpolation=_k4a.K4A_TRANSFORMATION_INTERPOLATION_TYPE_LINEAR,
    ):
        if not depth_image.is_valid() or not custom_image.is_valid():
            return Image()

        transformed_custom_image = Image.create(
            custom_image.format,
            self.color_resolution.width,
            self.color_resolution.height,
            self.color_resolution.width * self.get_custom_bytes_per_pixel(custom_image),
        )

        transformed_depth_image = Image.create(
            _k4a.K4A_IMAGE_FORMAT_DEPTH16,
            self.color_resolution.width,
            self.color_resolution.height,
            self.color_resolution.width * 2,
        )

        invalid_custom_value = ctypes.c_uint32()

        _k4a.k4a_transformation_depth_image_to_color_camera_custom(
            self._handle,
            depth_image.handle(),
            custom_image.handle(),
            transformed_depth_image.handle(),
            transformed_custom_image.handle(),
            interpolation,
            invalid_custom_value,
        )

        return transformed_custom_image

    def color_image_to_depth_camera(self, depth_image, color_image):
        if not depth_image.is_valid() or not color_image.is_valid():
            return Image()

        transformed_color_image = Image.create(
            _k4a.K4A_IMAGE_FORMAT_COLOR_BGRA32,
            self.depth_resolution.width,
            self.depth_resolution.height,
            self.depth_resolution.width * 4,
        )

        _k4a.k4a_transformation_color_image_to_depth_camera(
            self._handle,
            depth_image.handle(),
            color_image.handle(),
            transformed_color_image.handle(),
        )

        return transformed_color_image

    def depth_image_to_point_cloud(
        self, depth_image, calibration_type=_k4a.K4A_CALIBRATION_TYPE_DEPTH
    ):
        if not depth_image.is_valid():
            return Image()

        xyz_image = Image.create(
            _k4a.K4A_IMAGE_FORMAT_CUSTOM,
            depth_image.get_width_pixels(),
            depth_image.get_height_pixels(),
            depth_image.get_width_pixels() * 3 * 2,
        )

        _k4a.k4a_transformation_depth_image_to_point_cloud(
            self._handle, depth_image.handle(), calibration_type, xyz_image.handle()
        )

        return xyz_image

    def get_custom_bytes_per_pixel(self, custom_image):
        custom_image_format = custom_image.format

        if custom_image_format == _k4a.K4A_IMAGE_FORMAT_CUSTOM8:
            return 1
        else:
            return 2


class Resolution:
    def __init__(self, width, height):
        self.width = width
        self.height = height



from PySide6.QtCore import Slot
from PySide6.QtWidgets import (
    QFrame
)

class Asidebar(QFrame):
    def __init__(self) -> None:
        super().__init__()
        self.setStyleSheet("""
            background-color: black;
        """)
        # self.setFixedSize(320, 1000)
        self.hide()


import time
from PySide6.QtCore import Qt, Slot, QEvent, QMimeData
from PySide6.QtGui import QImage, QPixmap, QDrag
from PySide6.QtWidgets import (
    QHBoxLayout, QPushButton, QFrame, QGridLayout
)

from ..common_widgets import Frame
from .playback_sensors import PlaybackSensors
from .viewer_imu_sensors import ImuSensors
from .viewer_audio import AudioSensor
from pykinect_recorder.main._pyk4a.pykinect import initialize_libraries, start_playback
from ..signals import all_signals


SAMPLE_COUNT = 10000
RESOLUTION = 4


class PlaybackViewer(QFrame):
    def __init__(self, size: tuple[int, int] = (1200, 1000)) -> None:
        super().__init__()

        self.setStyleSheet("background-color: #1e1e1e;")
        self.th = None
        self.grid_layout = QGridLayout()
        self.frame_rgb = Frame("RGB Sensor")
        self.frame_depth = Frame("Depth Sensor")
        self.frame_ir = Frame("IR Sensor")
        
        self.layout_subdata = QHBoxLayout() # TODO => 네이밍 다시 하기
        self.imu_senser = ImuSensors()
        self.audio_sensor = AudioSensor()
        self.layout_subdata.addWidget(self.imu_senser)
        self.layout_subdata.addWidget(self.audio_sensor)
        self.frame_subdata = Frame("subdata", layout=self.layout_subdata)

        layout_btn = QHBoxLayout()
        layout_btn.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.btn_stop = QPushButton("Stop")
        self.btn_stop.setFixedSize(200, 40)
        self.btn_stop.setStyleSheet("""
            QPushButton:hover {
                border-color: "white";
            }
        """)
        layout_btn.addWidget(self.btn_stop)
        self.btn_stop.clicked.connect(self.stop_playback)
        
        self.target = None
        self.grid_layout.addWidget(self.frame_rgb, 0, 0)
        self.grid_layout.addWidget(self.frame_depth, 0, 1)
        self.grid_layout.addWidget(self.frame_ir, 1, 0)
        self.grid_layout.addWidget(self.frame_subdata, 1, 1)
        self.grid_layout.addLayout(layout_btn, 2, 0, 1, 2)

        self.setAcceptDrops(True)
        self.setLayout(self.grid_layout)

    def eventFilter(self, watched, event):
        if event.type() == QEvent.MouseButtonPress:
            self.mousePressEvent(event)
        elif event.type() == QEvent.MouseMove:
            self.mouseMoveEvent(event)
        elif event.type() == QEvent.MouseButtonRelease:
            self.mouseReleaseEvent(event)
        return super().eventFilter(watched, event)
    
    def get_index(self, pos):
        for i in range(self.grid_layout.count()):
            if self.grid_layout.itemAt(i).geometry().contains(pos) and i != self.target:
                return i

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.target = self.get_index(event.windowPos().toPoint())
        else:
            self.target = None

    def mouseMoveEvent(self, event):
        if event.buttons() & Qt.LeftButton and self.target is not None:
            drag = QDrag(self.grid_layout.itemAt(self.target).widget())
            pix = self.grid_layout.itemAt(self.target).widget().grab()
            mimedata = QMimeData()
            mimedata.setImageData(pix)
            drag.setMimeData(mimedata)
            drag.setPixmap(pix)
            drag.exec(Qt.DropAction.CopyAction | Qt.DropAction.MoveAction)

    def mouseReleaseEvent(self, event):
        self.target = None

    def dragEnterEvent(self, event):
        if event.mimeData().hasImage():
            event.acceptProposedAction()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        event.accept()

    def dropEvent(self, event):
        if not event.source().geometry().contains(event.pos()):
            source = self.get_index(event.pos())
            if source is None:
                return

            i, j = max(self.target, source), min(self.target, source)
            p1, p2 = self.grid_layout.getItemPosition(i), self.grid_layout.getItemPosition(j)

            self.grid_layout.addItem(self.grid_layout.takeAt(i), *p2)
            self.grid_layout.addItem(self.grid_layout.takeAt(j), *p1)

            event.accept()

    def stop_playback(self):
        self.th.is_run = False
        self.th.quit()

    @Slot(str)
    def playback(self, filepath) -> None:
        if self.th is not None and self.th.is_run:
            self.stop_playback()
            time.sleep(0.01)

        initialize_libraries()
        playback = start_playback(filepath)
        playback_config = playback.get_record_configuration()

        self.th = PlaybackSensors(playback=playback)
        all_signals.captured_rgb.connect(self.setRGBImage)
        all_signals.captured_depth.connect(self.setDepthImage)
        all_signals.captured_ir.connect(self.setIRImage)
        all_signals.captured_time.connect(self.setTime)
        all_signals.captured_fps.connect(self.setFps)

        # set option
        self.th.is_run = True
        self.th.start()
        
    @Slot(QImage)
    def setRGBImage(self, image: QImage) -> None:
        self.frame_rgb.frame.setPixmap(QPixmap.fromImage(image))
    
    @Slot(QImage)
    def setDepthImage(self, image: QImage) -> None:
        self.frame_depth.frame.setPixmap(QPixmap.fromImage(image))
        
    @Slot(QImage)
    def setIRImage(self, image: QImage) -> None:
        self.frame_ir.frame.setPixmap(QPixmap.fromImage(image))

    @Slot(float)
    def setTime(self, time) -> None:
        self.imu_senser.label_time.setText("Time(s) : %.3f" %time)

    @Slot(float)
    def setFps(self, value) -> None:
        self.imu_senser.label_fps.setText("FPS : %.2f" %value)



import ctypes


# K4A_DECLARE_HANDLE(k4a_device_t);
class _handle_k4a_device_t(ctypes.Structure):
    _fields_ = [
        ("_rsvd", ctypes.c_size_t),
    ]

# handling device open/close/start/capture
k4a_device_t = ctypes.POINTER(_handle_k4a_device_t)


# K4A_DECLARE_HANDLE(k4a_capture_t);
class _handle_k4a_capture_t(ctypes.Structure):
    _fields_ = [
        ("_rsvd", ctypes.c_size_t),
    ]

# after capturing image by device.open(), handling captured image
k4a_capture_t = ctypes.POINTER(_handle_k4a_capture_t)


# K4A_DECLARE_HANDLE(k4a_image_t);
class _handle_k4a_image_t(ctypes.Structure):
    _fields_ = [
        ("_rsvd", ctypes.c_size_t),
    ]

# handling image's metadata (exposure, h, w, size, timestamp, ...)
k4a_image_t = ctypes.POINTER(_handle_k4a_image_t)


# K4A_DECLARE_HANDLE(k4a_transformation_t);
class _handle_k4a_transformation_t(ctypes.Structure):
    _fields_ = [
        ("_rsvd", ctypes.c_size_t),
    ]

# format transformation like color_to_depth, depth_to_pts
k4a_transformation_t = ctypes.POINTER(_handle_k4a_transformation_t)

# class k4a_result_t(CtypeIntEnum):
k4a_result_t = ctypes.c_int
K4A_RESULT_SUCCEEDED = 0
K4A_RESULT_FAILED = 1

# class k4a_buffer_result_t(CtypeIntEnum):
k4a_buffer_result_t = ctypes.c_int
K4A_BUFFER_RESULT_SUCCEEDED = 0
K4A_BUFFER_RESULT_FAILED = 1
K4A_BUFFER_RESULT_TOO_SMALL = 2

# class k4a_wait_result_t(CtypeIntEnum):
k4a_wait_result_t = ctypes.c_int
K4A_WAIT_RESULT_SUCCEEDED = 0
K4A_WAIT_RESULT_FAILED = 1
K4A_WAIT_RESULT_TIMEOUT = 2

# Debug message level
# class k4a_log_level_t(CtypeIntEnum):
k4a_log_level_t = ctypes.c_int
K4A_LOG_LEVEL_CRITICAL = 0
K4A_LOG_LEVEL_ERROR = 1
K4A_LOG_LEVEL_WARNING = 2
K4A_LOG_LEVEL_INFO = 3
K4A_LOG_LEVEL_TRACE = 4
K4A_LOG_LEVEL_OFF = 5

# class k4a_depth_mode_t(CtypeIntEnum):
k4a_depth_mode_t = ctypes.c_int
K4A_DEPTH_MODE_OFF = 0
K4A_DEPTH_MODE_NFOV_2X2BINNED = 1
K4A_DEPTH_MODE_NFOV_UNBINNED = 2
K4A_DEPTH_MODE_WFOV_2X2BINNED = 3
K4A_DEPTH_MODE_WFOV_UNBINNED = 4
K4A_DEPTH_MODE_PASSIVE_IR = 5

# class k4a_color_resolution_t(CtypeIntEnum):
k4a_color_resolution_t = ctypes.c_int
K4A_COLOR_RESOLUTION_OFF = 0
K4A_COLOR_RESOLUTION_720P = 1
K4A_COLOR_RESOLUTION_1080P = 2
K4A_COLOR_RESOLUTION_1440P = 3
K4A_COLOR_RESOLUTION_1536P = 4
K4A_COLOR_RESOLUTION_2160P = 5
K4A_COLOR_RESOLUTION_3072P = 6

# class k4a_image_format_t(CtypeIntEnum):
k4a_image_format_t = ctypes.c_int
K4A_IMAGE_FORMAT_COLOR_MJPG = 0
K4A_IMAGE_FORMAT_COLOR_NV12 = 1
K4A_IMAGE_FORMAT_COLOR_YUY2 = 2
K4A_IMAGE_FORMAT_COLOR_BGRA32 = 3
K4A_IMAGE_FORMAT_DEPTH16 = 4
K4A_IMAGE_FORMAT_IR16 = 5
K4A_IMAGE_FORMAT_CUSTOM8 = 6
K4A_IMAGE_FORMAT_CUSTOM16 = 7
K4A_IMAGE_FORMAT_CUSTOM = 8

# depth_image to color interpolation 
# class k4a_transformation_interpolation_type_t(CtypeIntEnum):
k4a_transformation_interpolation_type_t = ctypes.c_int
K4A_TRANSFORMATION_INTERPOLATION_TYPE_NEAREST = 0
K4A_TRANSFORMATION_INTERPOLATION_TYPE_LINEAR = 1

# class k4a_fps_t(CtypeIntEnum):
k4a_fps_t = ctypes.c_int
K4A_FRAMES_PER_SECOND_5 = 0
K4A_FRAMES_PER_SECOND_15 = 1
K4A_FRAMES_PER_SECOND_30 = 2

# class k4a_color_control_command_t(CtypeIntEnum):
k4a_color_control_command_t = ctypes.c_int
K4A_COLOR_CONTROL_EXPOSURE_TIME_ABSOLUTE = 0
K4A_COLOR_CONTROL_AUTO_EXPOSURE_PRIORITY = 1
K4A_COLOR_CONTROL_BRIGHTNESS = 2
K4A_COLOR_CONTROL_CONTRAST = 3
K4A_COLOR_CONTROL_SATURATION = 4
K4A_COLOR_CONTROL_SHARPNESS = 5
K4A_COLOR_CONTROL_WHITEBALANCE = 6
K4A_COLOR_CONTROL_BACKLIGHT_COMPENSATION = 7
K4A_COLOR_CONTROL_GAIN = 8
K4A_COLOR_CONTROL_POWERLINE_FREQUENCY = 9

# class k4a_color_control_mode_t(CtypeIntEnum):
k4a_color_control_mode_t = ctypes.c_int
K4A_COLOR_CONTROL_MODE_AUTO = 0
K4A_COLOR_CONTROL_MODE_MANUAL = 1

# Using When record with two or more devices
# class k4a_wired_sync_mode_t(CtypeIntEnum):
k4a_wired_sync_mode_t = ctypes.c_int
K4A_WIRED_SYNC_MODE_STANDALONE = 0
K4A_WIRED_SYNC_MODE_MASTER = 1
K4A_WIRED_SYNC_MODE_SUBORDINATE = 2

# class k4a_calibration_type_t(CtypeIntEnum):
k4a_calibration_type_t = ctypes.c_int
K4A_CALIBRATION_TYPE_UNKNOWN = -1
K4A_CALIBRATION_TYPE_DEPTH = 0
K4A_CALIBRATION_TYPE_COLOR = 1
K4A_CALIBRATION_TYPE_GYRO = 2
K4A_CALIBRATION_TYPE_ACCEL = 3
K4A_CALIBRATION_TYPE_NUM = 4

# class k4a_calibration_model_type_t(CtypeIntEnum):
k4a_calibration_model_type_t = ctypes.c_int
K4A_CALIBRATION_LENS_DISTORTION_MODEL_UNKNOWN = 0
K4A_CALIBRATION_LENS_DISTORTION_MODEL_THETA = 1
K4A_CALIBRATION_LENS_DISTORTION_MODEL_POLYNOMIAL_3K = 2
K4A_CALIBRATION_LENS_DISTORTION_MODEL_RATIONAL_6KT = 3
K4A_CALIBRATION_LENS_DISTORTION_MODEL_BROWN_CONRADY = 4

# class k4a_firmware_build_t(CtypeIntEnum):
k4a_firmware_build_t = ctypes.c_int
K4A_FIRMWARE_BUILD_RELEASE = 0
K4A_FIRMWARE_BUILD_DEBUG = 1

# class k4a_firmware_signature_t(CtypeIntEnum):
k4a_firmware_signature_t = ctypes.c_int
K4A_FIRMWARE_SIGNATURE_MSFT = 0
K4A_FIRMWARE_SIGNATURE_TEST = 1
K4A_FIRMWARE_SIGNATURE_UNSIGNED = 2


# define K4A_SUCCEEDED(_result_) (_result_ == K4A_RESULT_SUCCEEDED)
def K4A_SUCCEEDED(result):
    return result == K4A_RESULT_SUCCEEDED


# define K4A_FAILED(_result_) (!K4A_SUCCEEDED(_result_))
def K4A_FAILED(result):
    return not K4A_SUCCEEDED(result)

# Callback Function
"""
typedef void(k4a_logging_message_cb_t)(
    void *context,
    k4a_log_level_t level,
    const char *file,
    const int line,
    const char *message
);

typedef void(k4a_memory_destroy_cb_t)(void *buffer, void *context);

typedef uint8_t *(k4a_memory_allocate_cb_t)(int size, void **context);
"""


class _k4a_device_configuration_t(ctypes.Structure):
    _fields_ = [
        ("color_format", ctypes.c_int),
        ("color_resolution", ctypes.c_int),
        ("depth_mode", ctypes.c_int),
        ("camera_fps", ctypes.c_int),
        ("synchronized_images_only", ctypes.c_bool),
        ("depth_delay_off_color_usec", ctypes.c_int32),
        ("wired_sync_mode", ctypes.c_int),
        ("subordinate_delay_off_master_usec", ctypes.c_uint32),
        ("disable_streaming_indicator", ctypes.c_bool),
    ]


k4a_device_configuration_t = _k4a_device_configuration_t


class _k4a_calibration_extrinsics_t(ctypes.Structure):
    _fields_ = [
        ("rotation", ctypes.c_float * 9),
        ("translation", ctypes.c_float * 3),
    ]

# rotation => 3x3(9) rotation, translation => Vector(x, y, z)
k4a_calibration_extrinsics_t = _k4a_calibration_extrinsics_t


class _param(ctypes.Structure):
    _fields_ = [
        ("cx", ctypes.c_float),
        ("cy", ctypes.c_float),
        ("fx", ctypes.c_float),
        ("fy", ctypes.c_float),
        ("k1", ctypes.c_float),
        ("k2", ctypes.c_float),
        ("k3", ctypes.c_float),
        ("k4", ctypes.c_float),
        ("k5", ctypes.c_float),
        ("k6", ctypes.c_float),
        ("codx", ctypes.c_float),
        ("cody", ctypes.c_float),
        ("p2", ctypes.c_float),
        ("p1", ctypes.c_float),
        ("metric_radius", ctypes.c_float),
    ]


class k4a_calibration_intrinsic_parameters_t(ctypes.Union):
    _fields_ = [
        ("param", _param),
        ("v", ctypes.c_float * 15),
    ]


class _k4a_calibration_intrinsics_t(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_int),
        ("parameter_count", ctypes.c_uint),
        ("parameters", k4a_calibration_intrinsic_parameters_t),
    ]


#v => Parameter counts (_params)
k4a_calibration_intrinsics_t = _k4a_calibration_intrinsics_t


class _k4a_calibration_camera_t(ctypes.Structure):
    _fields_ = [
        ("extrinsics", k4a_calibration_extrinsics_t),
        ("intrinsics", k4a_calibration_intrinsics_t),
        ("resolution_width", ctypes.c_int),
        ("resolution_height", ctypes.c_int),
        ("metric_radius", ctypes.c_float),
    ]


k4a_calibration_camera_t = _k4a_calibration_camera_t

# k4a_calibration_extrinsics_t[4][4] => 3D 변환을 가능하게 해줌.
class _k4a_calibration_t(ctypes.Structure):
    _fields_ = [
        ("depth_camera_calibration", k4a_calibration_camera_t),
        ("color_camera_calibration", k4a_calibration_camera_t),
        (
            "extrinsics",
            (k4a_calibration_extrinsics_t * K4A_CALIBRATION_TYPE_NUM)
            * K4A_CALIBRATION_TYPE_NUM,
        ),
        ("depth_mode", ctypes.c_int),
        ("color_resolution", ctypes.c_int),
    ]


k4a_calibration_t = _k4a_calibration_t


class _k4a_version_t(ctypes.Structure):
    _fields_ = [
        ("major", ctypes.c_uint32),
        ("minor", ctypes.c_uint32),
        ("iteration", ctypes.c_uint32),
    ]


k4a_version_t = _k4a_version_t


class _k4a_hardware_version_t(ctypes.Structure):
    _fields_ = [
        ("rgb", k4a_version_t),
        ("depth", k4a_version_t),
        ("audio", k4a_version_t),
        ("depth_sensor", k4a_version_t),
        ("firmware_build", ctypes.c_int),
        ("firmware_signature", ctypes.c_int),
    ]


k4a_hardware_version_t = _k4a_hardware_version_t


class _xy(ctypes.Structure):
    _fields_ = [
        ("x", ctypes.c_float),
        ("y", ctypes.c_float),
    ]

    def __iter__(self):
        return {"x": self.x, "y": self.y}

    def __str__(self):
        return str(self.__iter__())

# Two dimensional floating point vector
class k4a_float2_t(ctypes.Union):
    _fields_ = [("xy", _xy), ("v", ctypes.c_float * 2)]

    def __init__(self, v=(0, 0)):
        super().__init__()
        self.xy = _xy(v[0], v[1])

    def __iter__(self):
        xy = self.xy.__iter__()
        xy.update({"v": [v for v in self.v]})
        return xy

    def __str__(self):
        return self.xy.__str__()


class _xyz(ctypes.Structure):
    _fields_ = [
        ("x", ctypes.c_float),
        ("y", ctypes.c_float),
        ("z", ctypes.c_float),
    ]

    def __iter__(self):
        return {"x": self.x, "y": self.y, "z": self.z}

    def __str__(self):
        return str(self.__iter__())

# Three dimensional floating point vector
class k4a_float3_t(ctypes.Union):
    _fields_ = [("xyz", _xyz), ("v", ctypes.c_float * 3)]

    def __init__(self, v=(0, 0, 0)):
        super().__init__()
        self.xyz = _xyz(v[0], v[1], v[2])

    def __iter__(self):
        xyz = self.xyz.__iter__()
        xyz.update({"v": [v for v in self.v]})
        return xyz

    def __str__(self):
        return self.xyz.__str__()


class k4a_imu_sample_t(ctypes.Structure):
    _fields_ = [
        ("temperature", ctypes.c_float),
        ("acc_sample", k4a_float3_t),
        ("acc_timestamp_usec", ctypes.c_uint64),
        ("gyro_sample", k4a_float3_t),
        ("gyro_timestamp_usec", ctypes.c_uint64),
    ]


IMU_SAMPLE_SIZE = ctypes.sizeof(k4a_imu_sample_t)


K4A_DEVICE_DEFAULT = 0
K4A_WAIT_INFINITE = -1

color_command_dict = {
    "K4A_COLOR_CONTROL_EXPOSURE_TIME_ABSOLUTE": 0,
    "K4A_COLOR_CONTROL_AUTO_EXPOSURE_PRIORITY": 1,
    "K4A_COLOR_CONTROL_BRIGHTNESS": 2,
    "K4A_COLOR_CONTROL_CONTRAST": 3,
    "K4A_COLOR_CONTROL_SATURATION": 4,
    "K4A_COLOR_CONTROL_SHARPNESS": 5,
    "K4A_COLOR_CONTROL_WHITEBALANCE": 6,
    "K4A_COLOR_CONTROL_BACKLIGHT_COMPENSATION": 7,
    "K4A_COLOR_CONTROL_GAIN": 8,
    "K4A_COLOR_CONTROL_POWERLINE_FREQUENCY": 9
}


import numpy as np
import cv2

from pykinect_azure.k4abt import _k4abt
from pykinect_azure.k4abt.body import Body
from pykinect_azure.k4abt.body2d import Body2d
from pykinect_azure.k4abt._k4abtTypes import k4abt_body_t, body_colors
from pykinect_azure.k4a import Image, Capture, Transformation
from pykinect_azure.k4a._k4atypes import K4A_CALIBRATION_TYPE_DEPTH


class Frame:
    def __init__(self, frame_handle, calibration):
        if frame_handle:
            self._handle = frame_handle
            self.calibration = calibration
            self.transformation = Transformation(self.calibration)
            _k4abt.k4abt_frame_reference(self._handle)

    def __del__(self):
        self.reset()

    def json(self):
        bodies = self.get_bodies()

        if not bodies:
            return ""

        return [body.json() for body in bodies]

    def is_valid(self):
        return self._handle

    def handle(self):
        return self._handle

    def reset(self):
        if self.is_valid():
            self.release()
            self._handle = None

    def release(self):
        if self.is_valid():
            _k4abt.k4abt_frame_release(self._handle)

    def get_num_bodies(self):
        return _k4abt.k4abt_frame_get_num_bodies(self._handle)

    def get_body_skeleton(self, index=0):
        skeleton = _k4abt.k4abt_skeleton_t()

        _k4abt.VERIFY(
            _k4abt.k4abt_frame_get_body_skeleton(self._handle, index, skeleton),
            "Body tracker get body skeleton failed!",
        )

        return skeleton

    def get_body_id(self, index=0):
        return _k4abt.k4abt_frame_get_body_id(self._handle, index)

    def get_bodies(self):
        bodies = []

        # Get the number of people in the frame
        num_bodies = self.get_num_bodies()

        # Extract the skeleton of each person
        if num_bodies:
            for bodyIdx in range(num_bodies):
                bodies.append(self.get_body(bodyIdx))

        return bodies

    def get_body(self, bodyIdx=0):
        body_handle = k4abt_body_t()
        body_handle.id = self.get_body_id(bodyIdx)
        body_handle.skeleton = self.get_body_skeleton(bodyIdx)

        return Body(body_handle)

    def get_body2d(self, bodyIdx=0, dest_camera=K4A_CALIBRATION_TYPE_DEPTH):
        body_handle = self.get_body(bodyIdx).handle()

        return Body2d.create(body_handle, self.calibration, bodyIdx, dest_camera)

    def draw_bodies(
        self,
        destination_image,
        dest_camera=K4A_CALIBRATION_TYPE_DEPTH,
        only_segments=False,
    ):
        num_bodies = self.get_num_bodies()

        for body_id in range(num_bodies):
            destination_image = self.draw_body2d(
                destination_image, body_id, dest_camera, only_segments
            )

        return destination_image

    def draw_body2d(
        self,
        destination_image,
        bodyIdx=0,
        dest_camera=K4A_CALIBRATION_TYPE_DEPTH,
        only_segments=False,
    ):
        return self.get_body2d(bodyIdx, dest_camera).draw(
            destination_image, only_segments
        )

    def get_device_timestamp_usec(self):
        return _k4abt.k4abt_frame_get_device_timestamp_usec(self._handle)

    def get_body_index_map(self):
        return Image(_k4abt.k4abt_frame_get_body_index_map(self._handle))

    def get_body_index_map_image(self):
        return self.get_body_index_map().to_numpy()

    def get_transformed_body_index_map(self):
        depth_image = self.get_capture().get_depth_image_object()
        return self.transformation.depth_image_to_color_camera_custom(
            depth_image, self.get_body_index_map()
        )

    def get_transformed_body_index_map_image(self):
        transformed_body_index_map = self.get_transformed_body_index_map()
        return transformed_body_index_map.to_numpy()

    def get_segmentation_image(self):
        ret, body_index_map = self.get_body_index_map_image()
        return ret, np.dstack(
            [cv2.LUT(body_index_map, body_colors[:, i]) for i in range(3)]
        )

    def get_transformed_segmentation_image(self):
        ret, transformed_body_index_map = self.get_transformed_body_index_map_image()
        return ret, np.dstack(
            [cv2.LUT(transformed_body_index_map, body_colors[:, i]) for i in range(3)]
        )

    def get_capture(self):
        return Capture(
            _k4abt.k4abt_frame_get_capture(self._handle), self.calibration._handle
        )



import os

from PySide6.QtCore import Qt, Signal, Slot
from PySide6.QtGui import QIcon, QImage
from PySide6.QtWidgets import (
    QHBoxLayout, QMainWindow, QVBoxLayout, QWidget
)

from .renderer.components.toolbar import Toolbar
from .renderer.components.sidebar_menu import SidebarMenus
from .renderer.components.sidebar_control import StackedSidebar
from .renderer.components.viewer_control import StackedViewer
from .renderer.components.asidebar import Asidebar
from .renderer.signals import all_signals


class MainWindow(QMainWindow):
    def __init__(self) -> None:
        super().__init__()
        self.setWindowTitle("Azure Kinect Camera DK")
        self.setWindowIcon(QIcon(os.path.abspath("./renderer/public/kinect-sensor.ico")))
        self.initial_window()

    def initial_window(self) -> None:
        self.setFixedSize(1920, 1080)

        main_widget = QWidget()
        main_layout = QVBoxLayout(main_widget)
        main_layout.setSpacing(0)
        main_layout.setContentsMargins(0, 0, 0, 0)

        self.toolbar = Toolbar()
        main_layout.addWidget(self.toolbar)
        main_layout.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)

        main_sub_layout = QHBoxLayout()
        main_sub_layout.setSpacing(0)
        main_sub_layout.setContentsMargins(0, 0, 0, 0)
        main_sub_layout.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)

        self.sidebar_menus = SidebarMenus()
        self.stacked_sidebar = StackedSidebar()
        self.stacked_viewer = StackedViewer()
        self.asidebar = Asidebar()
        
        all_signals.save_filepath.connect(self.stacked_viewer.main_viewer.setBasePath)
        all_signals.config_viewer.connect(self.stacked_viewer.main_viewer.setConfig)
        all_signals.stacked_sidebar_status.connect(self.stacked_sidebar.setCurrentWidget)
        all_signals.stacked_sidebar_status.connect(self.stacked_viewer.setCurrentWidget)
        all_signals.playback_filepath.connect(self.stacked_viewer.main_explorer.playback)

        main_sub_layout.addWidget(self.sidebar_menus)
        main_sub_layout.addWidget(self.stacked_sidebar)
        main_sub_layout.addWidget(self.stacked_viewer)
        main_sub_layout.addWidget(self.asidebar)
        main_sub_layout.addStretch()

        main_layout.addLayout(main_sub_layout)
        self.setCentralWidget(main_widget)



import argparse
import os

from synology_api import downloadstation, filestation


def upload(args: argparse.ArgumentParser):
    try:
        f1 = filestation.FileStation(
            args.Synology_Ip,
            args.Synology_Port,
            args.Username,
            args.Password,
            args.secure,
            args.cert_verify,
            args.dsm_version,
            args.debug,
            args.otp_code,
        )

        success = f1.upload_file(
            args.dest_path, args.file_path, verify=args.cert_verify
        )
        if type(success) == tuple:
            raise ConnectionAbortedError
        else:
            print(success)

    except ValueError:
        print("Connect Failed!! Please Check your ip, port, username, password, secure")

    except FileNotFoundError:
        print("file path does not exists!!! Please Check before run")

    except ConnectionAbortedError:
        print("dest path does not match!!! Please Check before run")


def download(args: argparse.ArgumentParser):
    pass


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="Synology Nas Upload/Download",
        description="Uploading video and audio which is recoded by Azure Kinect",
    )

    ## Synology nas
    parser.add_argument(
        "--Synology_Ip",
        type=str,
        default="hnvlab.synology.me",
        help="Set Ip address such as xxx.synology.me or 111.111.1.1",
    )
    parser.add_argument("--Synology_Port", type=str, default="", help="")
    parser.add_argument("--Username", type=str, default="", help="")
    parser.add_argument("--Password", type=str, default="", help="")
    parser.add_argument(
        "--secure", type=bool, default=True, help="Set True if https is required"
    )
    parser.add_argument(
        "--cert_verify",
        type=bool,
        default=True,
        help="Set True if you want to verify your certificate",
    )
    parser.add_argument("--dsm_version", type=int, default=7)
    parser.add_argument("--debug", type=bool, default=True)
    parser.add_argument("--otp_code", type=str, default=None)

    parser.add_argument(
        "--dest_path",
        type=str,
        default="/dataset/test",
        help="Destination path in synology nas",
    )
    parser.add_argument(
        "--file_path", type=str, default="ecord_with_mic.py", help="Upload file path"
    )

    args = parser.parse_args()
    # upload(args)
    download(args)



import os
from typing import Tuple, Union, List
from PySide6.QtCore import Qt, QPoint
from PySide6.QtGui import QFont, QPen, QPainter, QFontMetrics

from PySide6.QtWidgets import (
    QLabel, QComboBox, QPushButton, QSlider, 
    QFrame, QVBoxLayout, QHBoxLayout
)


class ComboBox(QComboBox):
    def __init__(
        self, 
        items: List[str], 
        current_index: int,
        stylesheet: Union[str, os.PathLike] = None
    ) -> None:
        super().__init__()
        self.addItems(items)
        self.setCurrentIndex(current_index)

        # TODO: 외부 함수화 또는 PySide6에서 이런 함수 있는지 찾아보셈
        if stylesheet is not None:
            with open(os.path.join(os.path.split(__file__)[0], stylesheet), "r", encoding="utf-8") as f:
                stylesheet = f.read()
                print(stylesheet)
            self.setStyleSheet(str(stylesheet))


class PushButton(QPushButton):
    def __init__(
        self, 
        text: str = "",
        font: str = "Arial", 
        fontsize: int = 10, 
        stylesheet: Union[str, os.PathLike] = None
    ) -> None:
        super().__init__()
        self.setText(text)
        self.setFont(QFont(f"{font}", fontsize))
        self.setStyleSheet(
            "color: white;"
        )

        if stylesheet is not None:
            with open(os.path.join(os.path.split(__file__)[0], stylesheet), "r", encoding="utf-8") as f:
                stylesheet = f.read()
                print(stylesheet)
            self.setStyleSheet(str(stylesheet))


class Slider(QSlider):
    def __init__(
        self, 
        orientation, 
        set_range_values: Tuple[int], 
        set_value: int, 
    ) -> None:
        super().__init__(orientation)
        self.setRange(*set_range_values)
        self.setValue(set_value)

        self.setStyleSheet(
            """
            QSlider { 
                margin: 0px;
                border-radius: 4px;
            }
            QSlider::sub-page:horizontal {
                background-color: #3f4042;
                height: 12px;
                border-radius: 4px;
            }
            QSlider::groove:horizontal {
                height: 12px;
                margin: 1px;
                border-radius: 4px;
                background-color: "#3f4042"
            }
            QSlider::handle:horizontal {
                border: 10px;
                margin: 0px;
                border-radius: 3px;
                background-color: "#00bcf8";
            }
            QSlider:handle:horizontal:hover {
                background-color: "#4d96FF";
            }
            QSlider:handle:horizontal:pressed {
                background-color: "#FFFFFF";
            }
            """
        )

    def paintEvent(self, event):
        QSlider.paintEvent(self, event)

        curr_value = str(self.value())
        painter = QPainter(self)
        painter.setPen(QPen(Qt.white))

        font_metrics = QFontMetrics(self.font())
        font_width = font_metrics.boundingRect(curr_value).width()

        rect = self.geometry()
        if self.orientation() == Qt.Horizontal:
            horizontal_x_pos = rect.width()//2 - font_width//2 - 5
            horizontal_y_pos = rect.height() * 0.7
            painter.drawText(QPoint(horizontal_x_pos, horizontal_y_pos), curr_value)
        else:
            pass


class Label(QLabel):
    def __init__(
        self, 
        text: str = "",
        font: str = "Arial", 
        fontsize: int = 10, 
        orientation = None,
        stylesheet: Union[str, os.PathLike] = None
    ) -> None:
        super().__init__()
        self.setText(text)
        self.setFont(QFont(f"{font}", fontsize))
        if orientation is not None:
            self.setAlignment(orientation)

        if stylesheet is not None:
            with open(os.path.join(os.path.split(__file__)[0], stylesheet), "r", encoding="utf-8") as f:
                stylesheet = f.read()
                print(stylesheet)
            self.setStyleSheet(str(stylesheet))


class Frame(QFrame):
    def __init__(
        self,
        text: str,
        layout: Union[QVBoxLayout, QHBoxLayout] = None,
    ) -> None:
        super().__init__()
        self.setMaximumHeight(480)
        self.setFixedWidth(640)
        self.setObjectName("Frame")
        self.setStyleSheet(
            """QFrame#Frame {
                border-color: white;
            }"""
        )
        
        if layout is None:
            self.layout_main = QVBoxLayout()
            self.frame = Label(text, orientation=Qt.AlignmentFlag.AlignCenter)
            self.layout_main.addWidget(self.frame)
            self.setLayout(self.layout_main)
        else:
            self.setLayout(layout)


from . import _k4arecord
from .datablock import Datablock
from .record_configuration import RecordConfiguration
from ..k4a import _k4a
from ..k4a.capture import Capture
from ..k4a.calibration import Calibration
from ..k4a.imu_sample import ImuSample


class Playback:
    def __init__(self, filepath):
        self._handle = _k4arecord.k4a_playback_t()
        self._capture = None
        self._datablock = None

        self.open(filepath)
        self.calibration = self.get_calibration()

    def __del__(self):
        self.close()

    def open(self, filepath):
        _k4arecord.VERIFY(
            _k4arecord.k4a_playback_open(filepath.encode("utf-8"), self._handle),
            "Failed to open recording!",
        )

    def update(self):
        return self.get_next_capture()

    def is_valid(self):
        return self._handle != None

    def is_capture_initialized(self):
        return self._capture

    def is_datablock_initialized(self):
        return self._datablock

    def close(self):
        if self.is_valid():
            _k4arecord.k4a_playback_close(self._handle)
            self._handle = None

    def get_calibration(self):
        calibration_handle = _k4arecord.k4a_calibration_t()
        if self.is_valid():
            _k4arecord.VERIFY(
                _k4arecord.k4a_playback_get_calibration(
                    self._handle, calibration_handle
                ),
                "Failed to read device calibration from recording!",
            )

        return Calibration(calibration_handle)

    def get_record_configuration(self):
        config = _k4arecord.k4a_record_configuration_t()

        if self.is_valid():
            _k4arecord.VERIFY(
                _k4arecord.k4a_playback_get_record_configuration(self._handle, config),
                "Failed to read record configuration!",
            )

        return RecordConfiguration(config)

    def get_next_capture(self):
        capture_handle = _k4a.k4a_capture_t()

        if self.is_capture_initialized():
            self._capture.release_handle()
            self._capture._handle = capture_handle
        else:
            self._capture = Capture(capture_handle, self.calibration)

        ret = (
            _k4arecord.k4a_playback_get_next_capture(self._handle, capture_handle)
            != _k4arecord.K4A_STREAM_RESULT_EOF
        )

        return ret, self._capture

    def get_previous_capture(self):
        capture_handle = _k4a.k4a_capture_t()

        if self.is_capture_initialized():
            self._capture.release_handle()
            self._capture._handle = capture_handle
        else:
            self._capture = Capture(capture_handle, self.calibration)

        ret = (
            _k4arecord.k4a_playback_get_previous_capture(self._handle, capture_handle)
            != _k4arecord.K4A_STREAM_RESULT_EOF
        )

        return ret, self._capture

    def get_next_imu_sample(self):
        imu_sample_struct = _k4a.k4a_imu_sample_t()
        _k4a.VERIFY(
            _k4arecord.k4a_playback_get_next_imu_sample(
                self._handle, imu_sample_struct
            ),
            "Get next imu sample failed!",
        )

        # Convert the structure into a dictionary
        _imu_sample = ImuSample(imu_sample_struct)

        return _imu_sample

    def get_previous_imu_sample(self):
        imu_sample_struct = _k4a.k4a_imu_sample_t()
        _k4a.VERIFY(
            _k4arecord.k4a_playback_get_previous_imu_sample(
                self._handle, imu_sample_struct
            ),
            "Get previous imu sample failed!",
        )

        # Convert the structure into a dictionary
        _imu_sample = ImuSample(imu_sample_struct)

        return _imu_sample

    def seek_timestamp(self, offset=0, origin=_k4arecord.K4A_PLAYBACK_SEEK_BEGIN):
        _k4a.VERIFY(
            _k4arecord.k4a_playback_seek_timestamp(self._handle, offset, origin),
            "Seek recording failed!",
        )

    def get_recording_length(self):
        return int(_k4arecord.k4a_playback_get_recording_length_usec(self._handle))

    def set_color_conversion(self, format=_k4a.K4A_IMAGE_FORMAT_DEPTH16):
        _k4a.VERIFY(
            _k4arecord.k4a_playback_set_color_conversion(self._handle, format),
            "Seek color conversio failed!",
        )

    def get_next_data_block(self, track):
        block_handle = _k4arecord.k4a_playback_data_block_t()
        _k4a.VERIFY(
            _k4arecord.k4a_playback_get_next_data_block(
                self._handle, track, block_handle
            ),
            "Get next data block failed!",
        )

        if self.is_datablock_initialized():
            self._datablock._handle = block_handle
        else:
            self._datablock = Datablock(block_handle)

        return self._datablock

    def get_previous_data_block(self, track):
        block_handle = _k4arecord.k4a_playback_data_block_t()
        _k4a.VERIFY(
            _k4arecord.k4a_playback_get_previous_data_block(
                self._handle, track, block_handle
            ),
            "Get previous data block failed!",
        )

        if self.is_datablock_initialized():
            self._datablock._handle = block_handle
        else:
            self._datablock = Datablock(block_handle)

        return self._datablock



from . import _k4a


class Configuration:
    def __init__(self, configuration_handle=None):
        if configuration_handle:
            self._handle = configuration_handle
        else:
            self.create()

    def handle(self):
        return self._handle

    def __setattr__(self, name, value):
        """Run on change function when configuration parameters are changed"""

        if hasattr(self, name):
            if name != "_handle":
                if int(self.__dict__[name]) != value:
                    self.__dict__[name] = value
                    self.on_value_change()
            else:
                self.__dict__[name] = value
        else:
            self.__dict__[name] = value

    def __str__(self):
        """Print the current settings and a short explanation"""
        message = (
            "Device configuration: \n"
            f"\tcolor_format: {self.color_format} \n\t(0:JPG, 1:NV12, 2:YUY2, 3:BGRA32)\n\n"
            f"\tcolor_resolution: {self.color_resolution} \n\t(0:OFF, 1:720p, 2:1080p, 3:1440p, 4:1536p, 5:2160p, 6:3072p)\n\n"
            f"\tdepth_mode: {self.depth_mode} \n\t(0:OFF, 1:NFOV_2X2BINNED, 2:NFOV_UNBINNED,3:WFOV_2X2BINNED, 4:WFOV_UNBINNED, 5:Passive IR)\n\n"
            f"\tcamera_fps: {self.camera_fps} \n\t(0:5 FPS, 1:15 FPS, 2:30 FPS)\n\n"
            f"\tsynchronized_images_only: {self.synchronized_images_only} \n\t(True of False). Drop images if the color and depth are not synchronized\n\n"
            f"\tdepth_delay_off_color_usec: {self.depth_delay_off_color_usec} us. \n\tDelay between the color image and the depth image\n\n"
            f"\twired_sync_mode: {self.wired_sync_mode}\n\t(0:Standalone mode, 1:Master mode, 2:Subordinate mode)\n\n"
            f"\tsubordinate_delay_off_master_usec: {self.subordinate_delay_off_master_usec} us.\n\tThe external synchronization timing.\n\n"
            f"\tdisable_streaming_indicator: {self.disable_streaming_indicator} \n\t(True or False). Streaming indicator automatically turns on when the color or depth camera's are in use.\n\n"
        )
        return message

    def create(self):
        self.color_format = _k4a.K4A_IMAGE_FORMAT_COLOR_MJPG
        self.color_resolution = _k4a.K4A_COLOR_RESOLUTION_720P
        self.depth_mode = _k4a.K4A_DEPTH_MODE_NFOV_UNBINNED
        self.camera_fps = _k4a.K4A_FRAMES_PER_SECOND_30
        self.synchronized_images_only = False
        self.depth_delay_off_color_usec = 0
        self.wired_sync_mode = _k4a.K4A_WIRED_SYNC_MODE_STANDALONE
        self.subordinate_delay_off_master_usec = 0
        self.disable_streaming_indicator = False
        self.on_value_change()

    def create_from_handle(self, configuration_handle):
        self.color_format = configuration_handle.color_format
        self.color_resolution = configuration_handle.color_resolution
        self.depth_mode = configuration_handle.depth_mode
        self.camera_fps = configuration_handle.camera_fps
        self.synchronized_images_only = configuration_handle.synchronized_images_only
        self.depth_delay_off_color_usec = (
            configuration_handle.depth_delay_off_color_usec
        )
        self.wired_sync_mode = configuration_handle.wired_sync_mode
        self.subordinate_delay_off_master_usec = (
            configuration_handle.subordinate_delay_off_master_usec
        )
        self.disable_streaming_indicator = (
            configuration_handle.disable_streaming_indicator
        )

        self._handle = configuration_handle

    def on_value_change(self):
        self._handle = _k4a.k4a_device_configuration_t(
            self.color_format,
            self.color_resolution,
            self.depth_mode,
            self.camera_fps,
            self.synchronized_images_only,
            self.depth_delay_off_color_usec,
            self.wired_sync_mode,
            self.subordinate_delay_off_master_usec,
            self.disable_streaming_indicator,
        )


default_configuration = Configuration()



import os
import sys
import ctypes
import datetime
from pathlib import Path

from . import _k4a
from .capture import Capture
from .imu_sample import ImuSample
from .calibration import Calibration
from .configuration import Configuration
from ..k4arecord.record import Record
from ..k4a._k4atypes import K4A_WAIT_INFINITE


class Device:
    calibration = None
    capture = None
    imu_sample = None
    filename_video = None

    def __init__(self, index: int = 0) -> None:
        self._handle = None
        self._handle = self.open(index)
        self.recording = False
        self.record = False

    def __del__(self) -> None:
        self.close()

    def is_valid(self) -> None:
        return self._handle

    def is_capture_initialized(self) -> None:
        return Device.capture

    def is_imu_sample_initialized(self) -> None:
        return Device.imu_sample

    def handle(self) -> None:
        return self._handle

    def start(
        self, 
        configuration: Configuration,
        record=False,
        record_filepath="output.mkv"
    ) -> None:
        
        self.configuration = configuration
        self.start_cameras(configuration)
        self.start_imu()
        
        if record:
            self.record = Record(
                self._handle, self.configuration.handle(), record_filepath
            )
            self.recording = True

    def close(self) -> None:
        if self.is_valid():
            self.stop_imu()
            self.stop_cameras()
            _k4a.k4a_device_close(self._handle)

            # Clear members
            self._handle = None
            self.record = None
            self.recording = False

    def update(
        self, 
        timeout_in_ms: int = K4A_WAIT_INFINITE
    ) -> Capture:
        
        # Get cameras capture
        capture_handle = self.get_capture(timeout_in_ms)

        if self.is_capture_initialized():
            Device.capture._handle = capture_handle
        else:
            Device.capture = Capture(capture_handle, Device.calibration)

        # Write capture if recording
        if self.recording:
            self.record.write_capture(Device.capture.handle())

        return Device.capture

    def update_imu(
        self,
        timeout_in_ms: int = K4A_WAIT_INFINITE
    ) -> ImuSample:
        
        # Get imu sample
        imu_sample = self.get_imu_sample(timeout_in_ms)

        if self.is_imu_sample_initialized():
            Device.imu_sample._struct = imu_sample
            Device.imu_sample.parse_data()
        else:
            Device.imu_sample = ImuSample(imu_sample)

        return Device.imu_sample

    # TODO _k4a.K4A_WAIT_INFINITE는 없는데 ?
    def get_capture(
        self,
        timeout_in_ms: int = K4A_WAIT_INFINITE
    ) -> _k4a.ctypes.POINTER:
        
        # Release current handle
        if self.is_capture_initialized():
            Device.capture.release_handle()

        capture_handle = _k4a.k4a_capture_t()
        _k4a.VERIFY(
            _k4a.k4a_device_get_capture(self._handle, capture_handle, timeout_in_ms),
            "Get capture failed!",
        )

        return capture_handle

    def get_imu_sample(
        self,
        timeout_in_ms: int = K4A_WAIT_INFINITE
    ) -> _k4a.k4a_imu_sample_t:
        
        imu_sample = _k4a.k4a_imu_sample_t()

        _k4a.VERIFY(
            _k4a.k4a_device_get_imu_sample(self._handle, imu_sample, timeout_in_ms),
            "Get IMU failed!",
        )

        return imu_sample

    def start_cameras(
        self,
        device_config: Configuration
    ) -> None:

        Device.calibration = self.get_calibration(
            device_config.depth_mode, device_config.color_resolution
        )

        _k4a.VERIFY(
            _k4a.k4a_device_start_cameras(self._handle, device_config.handle()),
            "Start K4A cameras failed!",
        )

    def stop_cameras(self) -> None:
        _k4a.k4a_device_stop_cameras(self._handle)

    def start_imu(self) -> None:
        _k4a.VERIFY(_k4a.k4a_device_start_imu(self._handle), "Start K4A IMU failed!")

    def stop_imu(self) -> None:
        _k4a.k4a_device_stop_imu(self._handle)

    # get device serial number
    def get_serialnum(self) -> ctypes.c_int:
        
        serial_number_size = ctypes.c_size_t()
        result = _k4a.k4a_device_get_serialnum(self._handle, None, serial_number_size)

        if result == _k4a.K4A_BUFFER_RESULT_TOO_SMALL:
            serial_number = ctypes.create_string_buffer(serial_number_size.value)

        _k4a.VERIFY(
            _k4a.k4a_device_get_serialnum(
                self._handle, serial_number, serial_number_size
            ),
            "Read serial number failed!",
        )

        return serial_number.value.decode("utf-8")

    # ctypes.c_int => Configuration에 있는 Enum type
    def get_calibration(
        self,
        depth_mode: ctypes.c_int,
        color_resolution: ctypes.c_int
    ) -> Calibration:
        calibration_handle = _k4a.k4a_calibration_t()

        _k4a.VERIFY(
            _k4a.k4a_device_get_calibration(
                self._handle, depth_mode, color_resolution, calibration_handle
            ),
            "Get calibration failed!",
        )

        return Calibration(calibration_handle)

    def get_version(self):
        version = _k4a.k4a_hardware_version_t()

        _k4a.VERIFY(
            _k4a.k4a_device_get_version(self._handle, version), "Get version failed!"
        )

        return version

    @staticmethod
    def open(index=0):
        device_handle = _k4a.k4a_device_t()

        _k4a.VERIFY(
            _k4a.k4a_device_open(index, device_handle), "Open K4A Device failed!"
        )

        return device_handle

    @staticmethod
    def device_get_installed_count():
        return int(_k4a.k4a_device_get_installed_count())



import ctypes

from . import _k4a


class Calibration:
    def __init__(self, calibration_handle: _k4a.k4a_calibration_t):
        self._handle = calibration_handle
        self.color_params = (
            self._handle.color_camera_calibration.intrinsics.parameters.param
        )
        self.depth_params = (
            self._handle.depth_camera_calibration.intrinsics.parameters.param
        )

    def __del__(self):
        self.reset()

    def __str__(self):
        message = (
            "Rgb Intrinsic parameters: \n"
            f"\tcx: {self.color_params.cx}\n"
            f"\tcy: {self.color_params.cy}\n"
            f"\tfx: {self.color_params.fx}\n"
            f"\tfy: {self.color_params.fy}\n"
            f"\tk1: {self.color_params.k1}\n"
            f"\tk2: {self.color_params.k2}\n"
            f"\tk3: {self.color_params.k3}\n"
            f"\tk4: {self.color_params.k4}\n"
            f"\tk5: {self.color_params.k5}\n"
            f"\tk6: {self.color_params.k6}\n"
            f"\tcodx: {self.color_params.codx}\n"
            f"\tcody: {self.color_params.cody}\n"
            f"\tp2: {self.color_params.p2}\n"
            f"\tp1: {self.color_params.p1}\n"
            f"\tmetric_radius: {self.color_params.metric_radius}\n"
        )
        return message

    def get_matrix(self, camera: _k4a.k4a_calibration_type_t):
        if camera == _k4a.K4A_CALIBRATION_TYPE_COLOR:
            return [
                [self.color_params.fx, 0, self.color_params.cx],
                [0, self.color_params.fy, self.color_params.cy],
                [0, 0, 1],
            ]
        elif camera == _k4a.K4A_CALIBRATION_TYPE_DEPTH:
            return [
                [self.depth_params.fx, 0, self.depth_params.cx],
                [0, self.depth_params.fy, self.depth_params.cy],
                [0, 0, 1],
            ]

    def is_valid(self):
        return self._handle

    def handle(self):
        return self._handle

    def reset(self):
        if self.is_valid():
            self._handle = None

    # 3D point of source_camera to 3D point of target_camera  
    def convert_3d_to_3d(
        self,
        source_point3d: _k4a.k4a_float3_t(),
        source_camera: _k4a.k4a_calibration_type_t,
        target_camera: _k4a.k4a_calibration_type_t,
    ) -> _k4a.k4a_float3_t():
        target_point3d = _k4a.k4a_float3_t()

        _k4a.VERIFY(
            _k4a.k4a_calibration_3d_to_3d(
                self._handle,
                source_point3d,
                source_camera,
                target_camera,
                target_point3d,
            ),
            "Failed to convert from 3D to 3D",
        )

        return target_point3d

    # 2D depth of source_camera to 3D point of target_camera
    def convert_2d_to_3d(
        self,
        source_point2d: _k4a.k4a_float2_t,
        source_depth: float,
        source_camera: _k4a.k4a_calibration_type_t,
        target_camera: _k4a.k4a_calibration_type_t,
    ) -> _k4a.k4a_float3_t():
        target_point3d = _k4a.k4a_float3_t()
        valid = ctypes.c_int()

        _k4a.VERIFY(
            _k4a.k4a_calibration_2d_to_3d(
                self._handle,
                source_point2d,
                source_depth,
                source_camera,
                target_camera,
                target_point3d,
                valid,
            ),
            "Failed to convert from 2D to 3D",
        )

        return target_point3d

    # 3D point of source_camera to 2D pixel of target_camera
    def convert_3d_to_2d(
        self,
        source_point3d: _k4a.k4a_float3_t,
        source_camera: _k4a.k4a_calibration_type_t,
        target_camera: _k4a.k4a_calibration_type_t,
    ) -> _k4a.k4a_float2_t():
        target_point2d = _k4a.k4a_float2_t()
        valid = ctypes.c_int()

        _k4a.VERIFY(
            _k4a.k4a_calibration_3d_to_2d(
                self._handle,
                source_point3d,
                source_camera,
                target_camera,
                target_point2d,
                valid,
            ),
            "Failed to convert from 3D to 2D",
        )

        return target_point2d

    # 2D depth of source_camera to 2D pixel of target_camera
    def convert_2d_to_2d(
        self,
        source_point2d: _k4a.k4a_float2_t,
        source_depth: float,
        source_camera: _k4a.k4a_calibration_type_t,
        target_camera: _k4a.k4a_calibration_type_t,
    ) -> _k4a.k4a_float2_t():
        target_point2d = _k4a.k4a_float2_t()
        valid = ctypes.c_int()

        _k4a.VERIFY(
            _k4a.k4a_calibration_2d_to_2d(
                self._handle,
                source_point2d,
                source_depth,
                source_camera,
                target_camera,
                target_point2d,
                valid,
            ),
            "Failed to convert from 2D to 2D",
        )

        return target_point2d

    # 2D pixel of color_camera to 2D pixel of depth camera 
    def convert_color_2d_to_depth_2d(
        self, source_point2d: _k4a.k4a_float2_t, depth_image: _k4a.k4a_image_t
    ) -> _k4a.k4a_float2_t():
        target_point2d = _k4a.k4a_float2_t()
        valid = ctypes.c_int()

        _k4a.VERIFY(
            _k4a.k4a_calibration_color_2d_to_depth_2d(
                self._handle, source_point2d, depth_image, target_point2d, valid
            ),
            "Failed to convert from Color 2D to Depth 2D",
        )

        return target_point2d



import numpy as np
import cv2
import open3d as o3d


class Open3dVisualizer:
    def __init__(self):
        self.point_cloud = o3d.geometry.PointCloud()
        self.o3d_started = False

        self.vis = o3d.visualization.Visualizer()
        self.vis.create_window()

    def __call__(self, points_3d, rgb_image=None):
        self.update(points_3d, rgb_image)

    def update(self, points_3d, rgb_image=None):
        # Add values to vectors
        self.point_cloud.points = o3d.utility.Vector3dVector(points_3d)
        if rgb_image is not None:
            colors = cv2.cvtColor(rgb_image, cv2.COLOR_BGRA2RGB).reshape(-1, 3) / 255
            self.point_cloud.colors = o3d.utility.Vector3dVector(colors)

        self.point_cloud.transform(
            [[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]]
        )

        # Add geometries if it is the first time
        if not self.o3d_started:
            self.vis.add_geometry(self.point_cloud)
            self.o3d_started = True

        else:
            self.vis.update_geometry(self.point_cloud)

        self.vis.poll_events()
        self.vis.update_renderer()



import os
from pathlib import Path

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QHBoxLayout, QFrame, QFileDialog
)

from ..common_widgets import PushButton, Label
from ..signals import all_signals

class Toolbar(QFrame):
    def __init__(self) -> None:
        super().__init__()
        self.setMaximumWidth(1980)
        self.setMaximumHeight(50)
        self.setContentsMargins(0, 0, 0, 0)
        self.base_path = f"C:\\Users\\{os.getenv('USERNAME')}\\Videos\\"
        self.setObjectName("Toolbar")
        self.setStyleSheet("""
            QFrame#Toolbar {
                background-color: #323233; padding: 0px; margin: 0px;
            }
        """)

        self.btn_finddir = PushButton("경로 지정", "Arial", 10)
        self.btn_finddir.setFixedWidth(300)
        self.btn_finddir.clicked.connect(self.search_file)
        self.label_dirpath = Label(f"저장 경로 : {self.base_path}", orientation=Qt.AlignmentFlag.AlignLeft)

        btn_option = PushButton("옵션", "Arial", 10)
        btn_option.clicked.connect(self.option)

        layout_main = QHBoxLayout()
        layout_option = QHBoxLayout()
        layout_option.addWidget(btn_option)
        layout_option.setAlignment(Qt.AlignmentFlag.AlignRight)

        layout_main.addWidget(self.btn_finddir)
        layout_main.addWidget(self.label_dirpath)
        layout_main.addLayout(layout_option)
        self.setLayout(layout_main)

    def search_file(self) -> None:
        _dirNames = QFileDialog.getExistingDirectory(
            self, 
            "Open Data Files", 
            self.base_path,
            QFileDialog.ShowDirsOnly
        )
        self.label_dirpath.setText(_dirNames)
        self.base_path = _dirNames
        all_signals.save_filepath.emit(self.base_path)
        
    def option(self) -> None:
        pass



from PySide6.QtCore import Signal, QObject
from PySide6.QtGui import QImage

class AllSignals(QObject):
    # Stacked Widget signals
    stacked_sidebar_status = Signal(str)
    stacked_status = Signal(str)

    # Thread Signals
    captured_rgb = Signal(QImage)
    captured_depth = Signal(QImage)
    captured_ir = Signal(QImage)
    captured_time = Signal(float)
    captured_acc_data = Signal(list)
    captured_gyro_data = Signal(list)
    captured_fps = Signal(float)
    captured_audio = Signal(list)

    # playback/save_path Signals
    playback_filepath = Signal(str)
    save_filepath = Signal(str)
    is_run = Signal(bool)

    # config Signals
    config_viewer = Signal(dict)

    def __init__(self):
        super().__init__()
        pass
        

all_signals = AllSignals()


from pykinect_azure.k4abt import _k4abt
from pykinect_azure.k4abt.frame import Frame
from pykinect_azure.k4abt._k4abtTypes import k4abt_tracker_default_configuration
from pykinect_azure.k4a.device import Device
from pykinect_azure.k4a._k4atypes import K4A_WAIT_INFINITE
from pykinect_azure.utils import get_k4abt_lite_model_path


class Tracker:
    def __init__(self, calibration, model_type):
        self.calibration = calibration
        self._handle = self.create(model_type)
        self.frame = None

    def __del__(self):
        self.destroy()

    def is_valid(self):
        return self._handle

    def is_frame_initialized(self):
        return self.frame

    def handle(self):
        return self._handle

    def destroy(self):
        if self.is_valid():
            _k4abt.k4abt_tracker_destroy(self._handle)
            self._handle = None

    def update(self, capture=None, timeout_in_ms=K4A_WAIT_INFINITE):
        # Add capture to the body tracker processing queue
        if capture:
            self.enqueue_capture(capture.handle(), timeout_in_ms)
        else:
            self.enqueue_capture(Device.capture.handle(), timeout_in_ms)

        return self.pop_result(timeout_in_ms)

    def enqueue_capture(self, capture_handle, timeout_in_ms=K4A_WAIT_INFINITE):
        _k4abt.VERIFY(
            _k4abt.k4abt_tracker_enqueue_capture(
                self._handle, capture_handle, timeout_in_ms
            ),
            "Body tracker capture enqueue failed!",
        )

    def pop_result(self, timeout_in_ms=K4A_WAIT_INFINITE):
        if self.is_frame_initialized():
            self.frame.release()
            _k4abt.VERIFY(
                _k4abt.k4abt_tracker_pop_result(
                    self._handle, self.frame.handle(), timeout_in_ms
                ),
                "Body tracker get body frame failed!",
            )
        else:
            frame_handle = _k4abt.k4abt_frame_t()
            _k4abt.VERIFY(
                _k4abt.k4abt_tracker_pop_result(
                    self._handle, frame_handle, timeout_in_ms
                ),
                "Body tracker get body frame failed!",
            )
            self.frame = Frame(frame_handle, self.calibration)

        return self.frame

    def set_temporal_smoothing(self, smoothing_factor):
        _k4abt.k4abt_tracker_set_temporal_smoothing(self._handle, smoothing_factor)

    def shutdown(self):
        _k4abt.k4abt_tracker_shutdown(self._handle)

    def create(self, model_type):
        tracker_config = self.get_tracker_configuration(model_type)

        tracker_handle = _k4abt.k4abt_tracker_t()
        _k4abt.VERIFY(
            _k4abt.k4abt_tracker_create(
                self.calibration.handle(), tracker_config, tracker_handle
            ),
            "Body tracker initialization failed!",
        )

        return tracker_handle

    def get_tracker_configuration(self, model_type):
        tracker_config = k4abt_tracker_default_configuration

        if model_type == _k4abt.K4ABT_LITE_MODEL:
            tracker_config.model_path = get_k4abt_lite_model_path()

        return tracker_config



import os
from PySide6.QtCore import QSize, Qt
from PySide6.QtWidgets import (
    QVBoxLayout, QFrame, QPushButton
)

import qtawesome as qta
from ..signals import all_signals

class SidebarMenus(QFrame):
    def __init__(self) -> None:
        super().__init__()
        self.setFixedWidth(60)
        self.setStyleSheet("background-color: #333333;")
        _root_path = os.path.abspath("./renderer/public/")

        main_layout = QVBoxLayout()
        main_layout.setAlignment(Qt.AlignTop | Qt.AlignCenter)

        self.recorder_option = self._make_icons(
            qta.icon("mdi.file-multiple-outline"),
            "Recording Viewer",
        )
        self.explorer_option = self._make_icons(
            qta.icon("mdi6.file-find-outline"),
            "Explorer & Playback",
        )
        self.deeplearning_option = self._make_icons( 
            qta.icon("ph.gear"), 
            "Deep Learning Solution",
        )

        main_layout.addWidget(self.recorder_option)
        main_layout.addWidget(self.explorer_option)
        main_layout.addWidget(self.deeplearning_option)
        self.setLayout(main_layout)

        self.recorder_option.clicked.connect(self.clicked_recorder)
        self.explorer_option.clicked.connect(self.clicked_explorer)
        self.deeplearning_option.clicked.connect(self.clicked_solution)

    def _make_icons(
            self, 
            icon: qta,
            tooltip: str,
        ) -> QPushButton:

        _btn = QPushButton(icon, "")
        _btn.setFixedSize(45, 55)
        _btn.setIconSize(QSize(45, 45))
        _btn.setToolTip(f'<b>{tooltip}<b>')
        _btn.setStyleSheet("""
            QPushButton:hover {
                border-color: white;
            }
            QToolTip {
                font:"Arial"; font-size: 15px; color: #ffffff; border: 1px solid #ffffff; 
            }
        """)

        # print(root_path)
        # print(os.path.join(root_path, file_path))
        # tmp = QPixmap(os.path.join(root_path, file_path))
        # tmp = QPixmap(os.path.join(os.path.split(__file__)[0], file_path))
        # _icon = QIcon(tmp)
        # _btn.setIcon(QIcon(os.path.join(root_path, file_path)))
        # _btn.setIcon(file_path)

        # print(os.path.abspath("./renderer/components/sidebarBtn.stylesheet"))
        # print(os.path.split(__file__))
        # if stylesheet is not None:
        #     with open(os.path.join(os.path.split(__file__)[0], stylesheet), "r", encoding="utf-8") as f:
        #     # with open(os.path.abspath("./pykinect_recorder/renderer/components/sidebbarBtn.stylesheet"), "r", encoding="utf-8") as f:
        #         stylesheet = f.read()
        #     _btn.setStyleSheet(str(stylesheet))

        return _btn
    
    def clicked_recorder(self):
        all_signals.stacked_sidebar_status.emit("recorder")

    def clicked_explorer(self):
        all_signals.stacked_sidebar_status.emit("explorer")

    def clicked_solution(self):
        all_signals.stacked_sidebar_status.emit("solution")



import os
import sys
import time
import ctypes
import datetime
from pathlib import Path

from PySide6.QtCore import Qt, Slot, QEvent, QMimeData, QPointF
from PySide6.QtGui import QImage, QPixmap, QDrag
from PySide6.QtWidgets import (
    QHBoxLayout, QPushButton, QVBoxLayout, 
    QFrame, QDialog, QGridLayout
)

from ..common_widgets import Label, Frame
from .record_sensors import RecordSensors
from .playback_sensors import PlaybackSensors
from .viewer_imu_sensors import ImuSensors
from .viewer_audio import AudioSensor
from pykinect_recorder.main._pyk4a.k4a._k4a import k4a_device_set_color_control
from pykinect_recorder.main._pyk4a.k4a._k4atypes import color_command_dict, K4A_COLOR_CONTROL_MODE_MANUAL
from pykinect_recorder.main._pyk4a.k4a.configuration import Configuration
from pykinect_recorder.main._pyk4a.pykinect import start_device, initialize_libraries, start_playback
from .sidebar_record_control import _config_sidebar
from ..signals import all_signals


SAMPLE_COUNT = 10000
RESOLUTION = 4


class SensorViewer(QFrame):
    def __init__(self, size: tuple[int, int] = (1200, 1000)) -> None:
        super().__init__()

        self.setStyleSheet("background-color: #1e1e1e;") 
        self.device = None
        self.config = None
        self.color_control = None
        self.base_path = None
        self.emit_configs = _config_sidebar

        self.grid_layout = QGridLayout()
        self.frame_rgb = Frame("RGB Sensor")
        self.frame_depth = Frame("Depth Sensor")
        self.frame_ir = Frame("IR Sensor")
        
        self.layout_subdata = QHBoxLayout() # TODO => 네이밍 다시 하기
        self.imu_senser = ImuSensors()
        self.audio_sensor = AudioSensor()
        self.layout_subdata.addWidget(self.imu_senser)
        self.layout_subdata.addWidget(self.audio_sensor)
        self.frame_subdata = Frame("subdata", layout=self.layout_subdata)

        layout_btn = QHBoxLayout()
        layout_btn.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.btn_open = QPushButton("Device open")
        self.btn_open.setFixedSize(200, 40)
        self.btn_viewer = QPushButton("▶")
        self.btn_viewer.setFixedSize(200, 40)
        self.btn_record = QPushButton("●")
        self.btn_record.setFixedSize(200, 40)

        self.btn_open.setStyleSheet("""
            QPushButton:hover {
                border-color: "white";
            }
        """)
        self.btn_viewer.setStyleSheet("""
            QToolTip {
                font:"Arial"; font-size: 15px; color: #ffffff; border: 1px solid #ffffff; 
            }
            QPushButton:hover {
                border-color: "white";
            }
            """
        )
        self.btn_record.setStyleSheet("""
            QToolTip {
                font:"Arial"; font-size: 15px; color: #ffffff; border: 1px solid #ffffff; 
            }
            QPushButton {
                background-color: red;
            }
            QPushButton:hover {
                border-color: "white";
            }
            """
        )
        

        self.btn_viewer.setToolTip("<b>Streaming Button</b>")
        self.btn_record.setToolTip("<b>Recording Button</b>")

        layout_btn.addStretch()
        layout_btn.addStretch()
        layout_btn.addWidget(self.btn_open)
        layout_btn.addStretch()
        layout_btn.addWidget(self.btn_viewer)
        layout_btn.addStretch()
        layout_btn.addWidget(self.btn_record)      
        layout_btn.addStretch()
        layout_btn.addStretch()
        
        self.buffer = [QPointF(x, 0) for x in range(SAMPLE_COUNT)]
        self.th = RecordSensors(device=self.device)
        all_signals.captured_rgb.connect(self.setRGBImage)
        all_signals.captured_depth.connect(self.setDepthImage)
        all_signals.captured_ir.connect(self.setIRImage)
        all_signals.captured_time.connect(self.setTime)
        all_signals.captured_acc_data.connect(self.setAccData)
        all_signals.captured_gyro_data.connect(self.setGyroData)
        all_signals.captured_fps.connect(self.setFps)
        all_signals.captured_audio.connect(self.setAudioData)
            
        self.is_device = True
        self.is_viewer = True
        self.is_record = True
        self.btn_open.clicked.connect(self.open_device)
        self.btn_viewer.clicked.connect(self.streaming)
        self.btn_record.clicked.connect(self.recording)
        self.btn_viewer.setEnabled(False)
        
        self.target = None
        self.grid_layout.addWidget(self.frame_rgb, 0, 0)
        self.grid_layout.addWidget(self.frame_depth, 0, 1)
        self.grid_layout.addWidget(self.frame_ir, 1, 0)
        self.grid_layout.addWidget(self.frame_subdata, 1, 1)
        self.grid_layout.addLayout(layout_btn, 2, 0, 1, 2)

        self.setAcceptDrops(True)
        self.setLayout(self.grid_layout)

    def eventFilter(self, watched, event):
        if event.type() == QEvent.MouseButtonPress:
            self.mousePressEvent(event)
        elif event.type() == QEvent.MouseMove:
            self.mouseMoveEvent(event)
        elif event.type() == QEvent.MouseButtonRelease:
            self.mouseReleaseEvent(event)
        return super().eventFilter(watched, event)
    
    def get_index(self, pos):
        for i in range(self.grid_layout.count()):
            if self.grid_layout.itemAt(i).geometry().contains(pos) and i != self.target:
                return i

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.target = self.get_index(event.windowPos().toPoint())
        else:
            self.target = None

    def mouseMoveEvent(self, event):
        if event.buttons() & Qt.LeftButton and self.target is not None:
            drag = QDrag(self.grid_layout.itemAt(self.target).widget())
            pix = self.grid_layout.itemAt(self.target).widget().grab()
            mimedata = QMimeData()
            mimedata.setImageData(pix)
            drag.setMimeData(mimedata)
            drag.setPixmap(pix)
            drag.exec(Qt.DropAction.CopyAction | Qt.DropAction.MoveAction)

    def mouseReleaseEvent(self, event):
        self.target = None

    def dragEnterEvent(self, event):
        if event.mimeData().hasImage():
            event.acceptProposedAction()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        event.accept()

    def dropEvent(self, event):
        if not event.source().geometry().contains(event.pos()):
            source = self.get_index(event.pos())
            if source is None:
                return

            i, j = max(self.target, source), min(self.target, source)
            p1, p2 = self.grid_layout.getItemPosition(i), self.grid_layout.getItemPosition(j)

            self.grid_layout.addItem(self.grid_layout.takeAt(i), *p2)
            self.grid_layout.addItem(self.grid_layout.takeAt(j), *p1)

            event.accept()
        
    def check_device(self) -> bool:
        try:
            self.config = Configuration()     
            initialize_libraries()
            _device = start_device(config=self.config)
            _device.close()
        except:
            modal = QDialog()
            layout_modal = QVBoxLayout()
            e_message = Label(
                "<b>카메라 연결에 문제가 있습니다. <br> 연결을 재시도해주세요.</b>", 
                "Arial", 20, Qt.AlignmentFlag.AlignCenter
            )
            layout_modal.addWidget(e_message)
            modal.setLayout(layout_modal)
            modal.setWindowTitle("Error Message")
            modal.resize(400, 200)
            modal.exec()

    def open_device(self) -> None:
        if self.is_device is True:
            self.check_device()
            self.is_device = False
            self.btn_open.setText("Device close")
        else:
            self.frame_rgb.frame.setText("RGB Frame")
            self.frame_depth.frame.setText("Depth Frame")
            self.frame_ir.frame.setText("IR Frame")
            self.is_device = True
            self.btn_open.setText("Device open")
            self.device = None
    
    # TODO Streaming 이랑 Recording 겹치는 코드가 많음.
    def streaming(self) -> None:
        if self.is_viewer:
            self.set_filename()

            for k, v in self.emit_configs["color"].items():
                setattr(self.config, k, v)
            setattr(self.config, "depth_mode", self.emit_configs["depth_mode"])
            self.device = start_device(config=self.config, record=False)   

            for k, v in self.emit_configs["color_option"].items():
                k4a_device_set_color_control(
                    self.device._handle,
                    color_command_dict[k],
                    K4A_COLOR_CONTROL_MODE_MANUAL,
                    ctypes.c_int32(int(v))
                )

            self.th.device = self.device
            self.th.audio_file = self.filename_audio

            self.btn_record.setEnabled(False)
            self.btn_open.setEnabled(False)
            self.btn_viewer.setText("■")
            self.th.is_run = True
            self.is_viewer = False
            self.th.start()
        else:
            self.btn_record.setEnabled(True)
            self.btn_open.setEnabled(True)
            self.btn_viewer.setText("▶")
            self.th.is_run = False
            self.is_viewer = True
            self.device.close()
            self.th.quit()
            time.sleep(1)   
        
    def recording(self) -> None:
        if self.is_record:
            self.set_filename()

            for k, v in self.emit_configs["color"].items():
                setattr(self.config, k, v)
            setattr(self.config, "depth_mode", self.emit_configs["depth_mode"])
            self.device = start_device(
                config=self.config, 
                record_filepath=self.filename_video,
                record=True
            )   

            for k, v in self.emit_configs["color_option"].items():
                k4a_device_set_color_control(
                    self.device._handle,
                    color_command_dict[k],
                    K4A_COLOR_CONTROL_MODE_MANUAL,
                    ctypes.c_int32(int(v))
                )
            
            self.th.device = self.device
            self.th.audio_record = True
            self.th.audio_file = self.filename_audio

            # self.btn_viewer.setEnabled(False)
            self.btn_open.setEnabled(False)
            self.btn_record.setText("■")
            self.th.is_run = True
            self.is_record = False
            self.th.start()
        else:
            # self.btn_viewer.setEnabled(True)
            self.btn_open.setEnabled(True)
            self.btn_record.setText("▶")
            self.th.is_run = False
            self.is_record = True
            self.device.close()
            self.th.quit()
            time.sleep(1)

    def set_filename(self) -> None:
        if self.base_path is None:
            self.base_path = os.path.join(Path.home(), "Videos")

        filename = datetime.datetime.now()
        filename = filename.strftime("%Y_%m_%d_%H_%M_%S")

        self.filename_video = os.path.join(self.base_path, f"{filename}.mkv")
        self.filename_audio = os.path.join(self.base_path, f"{filename}.mp3")
        if sys.flags.debug:
            print(self.base_path, self.filename_video)

    @Slot(dict)
    def setConfig(self, value: dict) -> None:
        self.emit_configs = value

    @Slot(str)
    def setBasePath(self, value: str) -> None:
        self.base_path = value
        
    @Slot(QImage)
    def setRGBImage(self, image: QImage) -> None:
        self.frame_rgb.frame.setPixmap(QPixmap.fromImage(image))
    
    @Slot(QImage)
    def setDepthImage(self, image: QImage) -> None:
        self.frame_depth.frame.setPixmap(QPixmap.fromImage(image))
        
    @Slot(QImage)
    def setIRImage(self, image: QImage) -> None:
        self.frame_ir.frame.setPixmap(QPixmap.fromImage(image))

    @Slot(float)
    def setTime(self, time) -> None:
        self.imu_senser.label_time.setText("Time(s) : %.3f" %time)

    @Slot(float)
    def setFps(self, value) -> None:
        self.imu_senser.label_fps.setText("FPS : %.2f" %value)

    @Slot(list)
    def setAccData(self, values) -> None:
        self.imu_senser.acc_x.setText("X : %.5f" %values[0])
        self.imu_senser.acc_y.setText("Y : %.5f" %values[1])
        self.imu_senser.acc_z.setText("Z : %.5f" %values[2])

    @Slot(float)
    def setGyroData(self, values) -> None:
        self.imu_senser.gyro_x.setText("X : %.5f" %values[0])
        self.imu_senser.gyro_y.setText("Y : %.5f" %values[1])
        self.imu_senser.gyro_z.setText("Z : %.5f" %values[2])

    @Slot(list)
    def setAudioData(self, values) -> None:
        start = 0
        if (values[1] < SAMPLE_COUNT):
            start = SAMPLE_COUNT - values[1]
            for s in range(start):
                self.buffer[s].setY(self.buffer[s + values[1]].y())

        data_index = 0
        for s in range(start, SAMPLE_COUNT):
            value = (ord(values[0][data_index]) - 128) / 128
            self.buffer[s].setY(value)
            data_index = data_index + RESOLUTION
        
        self.audio_sensor.series.replace(self.buffer)


from PySide6.QtCore import Slot
from PySide6.QtWidgets import (
    QStackedLayout, QFrame
)

from .sidebar_solutions import SolutionSidebar
from .sidebar_record_control import ViewerSidebar
from .sidebar_explorer import ExplorerSidebar


class StackedSidebar(QFrame):
    def __init__(self) -> None:
        super().__init__()

        self.main_layout = QStackedLayout()
        self.sidebar_viewer = ViewerSidebar()
        self.sidebar_explorer = ExplorerSidebar()
        self.sidebar_solution = SolutionSidebar()
        
        self.main_layout.addWidget(self.sidebar_viewer) 
        self.main_layout.addWidget(self.sidebar_explorer) 
        self.main_layout.addWidget(self.sidebar_solution) 
        
        # 현재 index
        self.main_layout.setCurrentIndex(0)
        self.setLayout(self.main_layout)


    @Slot(str)
    def setCurrentWidget(self, value):
        if value == "explorer":
            self.main_layout.setCurrentWidget(self.sidebar_explorer)
        elif value == "solution":
            self.main_layout.setCurrentWidget(self.sidebar_solution)
        else:
            self.main_layout.setCurrentWidget(self.sidebar_viewer)


from . import _k4arecord
from ..k4a import _k4a


class Datablock:
    def __init__(self, modulePath):
        self._handle = _k4arecord.k4a_playback_data_block_t()

    def __del__(self):
        self.reset()

    def is_valid(self):
        return self.datablock_handle != None

    def handle(self):
        return self._handle

    def reset(self):
        if self.is_valid():
            _k4arecord.k4a_playback_data_block_release(self._handle)
            self._handle = None

    def get_device_timestamp_usec(self):
        return int(
            _k4arecord.k4a_playback_data_block_get_device_timestamp_usec(self._handle)
        )

    def get_buffer_size(self):
        return int(_k4arecord.k4a_playback_data_block_get_buffer_size(self._handle))

    def get_buffer(self):
        if not self.is_valid():
            return None

        return _k4arecord.k4a_playback_data_block_get_buffer(self._handle)



import cv2

from . import _k4a
from .image import Image
from .transformation import Transformation
from ..utils.postProcessing import smooth_depth_image


class Capture:
    def __init__(self, capture_handle, calibration):
        self._handle = capture_handle
        self.calibration = calibration
        self.camera_transform = Transformation(calibration)

    def __del__(self):
        self.reset()

    def is_valid(self):
        return self._handle

    def handle(self):
        return self._handle

    def reset(self):
        if self.is_valid():
            self.release_handle()
            self._handle = None

    def release_handle(self):
        if self.is_valid():
            _k4a.k4a_capture_release(self._handle)

    @staticmethod
    def create():
        handle = _k4a.k4a_capture_t
        _k4a.VERIFY(Capture._k4a.k4a_capture_create(handle), "Create capture failed!")

        return Capture(handle)

    def get_color_image_object(self):
        return Image(_k4a.k4a_capture_get_color_image(self._handle))

    def get_depth_image_object(self):
        return Image(_k4a.k4a_capture_get_depth_image(self._handle))

    def get_ir_image_object(self):
        return Image(_k4a.k4a_capture_get_ir_image(self._handle))

    def get_transformed_depth_object(self):
        return self.camera_transform.depth_image_to_color_camera(
            self.get_depth_image_object()
        )

    def get_transformed_color_object(self):
        return self.camera_transform.color_image_to_depth_camera(
            self.get_depth_image_object(), self.get_color_image_object()
        )

    def get_pointcloud_object(self, calibration_type=_k4a.K4A_CALIBRATION_TYPE_DEPTH):
        return self.camera_transform.depth_image_to_point_cloud(
            self.get_depth_image_object(), calibration_type
        )

    def get_transformed_pointcloud_object(self):
        return self.camera_transform.depth_image_to_point_cloud(
            self.get_transformed_depth_object(), _k4a.K4A_CALIBRATION_TYPE_COLOR
        )

    def get_color_image(self):
        return self.get_color_image_object().to_numpy()

    def get_depth_image(self):
        return self.get_depth_image_object().to_numpy()

    def get_colored_depth_image(self):
        ret, depth_image = self.get_depth_image()
        if not ret:
            return ret, None

        return ret, self.color_depth_image(depth_image)

    def get_ir_image(self):
        return self.get_ir_image_object().to_numpy()

    def get_transformed_depth_image(self):
        return self.get_transformed_depth_object().to_numpy()

    def get_transformed_colored_depth_image(self):
        ret, transformed_depth_image = self.get_transformed_depth_image()

        return ret, self.color_depth_image(transformed_depth_image)

    def get_transformed_color_image(self):
        return self.get_transformed_color_object().to_numpy()

    def get_smooth_depth_image(self, maximum_hole_size=10):
        ret, depth_image = self.get_depth_image()
        return ret, smooth_depth_image(depth_image, maximum_hole_size)

    def get_smooth_colored_depth_image(self, maximum_hole_size=10):
        ret, smooth_depth_image = self.get_smooth_depth_image(maximum_hole_size)
        return ret, self.color_depth_image(smooth_depth_image)

    def get_pointcloud(self, calibration_type=_k4a.K4A_CALIBRATION_TYPE_DEPTH):
        ret, points = self.get_pointcloud_object(calibration_type).to_numpy()
        points = points.reshape((-1, 3))
        return ret, points

    def get_transformed_pointcloud(self):
        ret, points = self.get_transformed_pointcloud_object().to_numpy()
        points = points.reshape((-1, 3))
        return ret, points

    @staticmethod
    def color_depth_image(depth_image):
        depth_color_image = cv2.convertScaleAbs(
            depth_image, alpha=0.05
        )  # alpha is fitted by visual comparison with Azure k4aviewer results
        depth_color_image = cv2.applyColorMap(depth_color_image, cv2.COLORMAP_JET)

        return depth_color_image



import argparse
import datetime
import os
import queue
import sys
import tempfile

import numpy  # Make sure NumPy is loaded before it is used in the callback

import sounddevice as sd
import soundfile as sf

assert numpy  # avoid "imported but unused" message (W0611)

import os
import sys
import time
from argparse import ArgumentParser

import cv2
import numpy as np
import pyk4a
from pydub import AudioSegment
from pyk4a import Config, ImageFormat, PyK4A, PyK4ARecord
from PySide6.QtCore import Qt, QThread, Signal, Slot
from PySide6.QtGui import QAction, QImage, QKeySequence, QPixmap
from PySide6.QtWidgets import (
    QApplication,
    QComboBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)


if not os.path.exists("./tools/pyk4a/example/outputs/"):
    os.mkdir("./tools/pyk4a/example/outputs")


def callback(indata, frames, time, status):
    global q
    """This is called (from a separate thread) for each audio block."""
    if status:
        print(status, file=sys.stderr)
    q.put(indata.copy())


class Thread(QThread):
    updateFrame = Signal(QImage)

    def __init__(self, parent=None):
        QThread.__init__(self, parent)
        self.trained_file = None
        self.status = True
        self.cap = True
        # self.args = args

        self.set_filename()

    def set_filename(self):
        filename = datetime.datetime.now()
        filename = filename.strftime("%Y_%m_%d_%H_%M_%S")
        base_path = (
            "C:\\Program Files\\Azure Kinect SDK v1.4.1\\tools\\pyk4a\\example\\outputs"
        )

        self.filename_video = f"{base_path}\\{filename}.mkv"
        self.filename_audio = f"{base_path}\\{filename}.wav"
        print(filename)

    def run(self):
        global q
        # pyk4a
        config = Config(
            color_format=ImageFormat.COLOR_BGRA32,
            depth_mode=pyk4a.DepthMode.NFOV_UNBINNED,
            color_resolution=pyk4a.ColorResolution.RES_720P,
            synchronized_images_only=True,
        )
        azure_device = PyK4A(config=config, device_id=0)
        azure_device.start()
        record = PyK4ARecord(
            device=azure_device, config=config, path=self.filename_video
        )
        record.create()

        with sf.SoundFile(
            self.filename_audio,
            mode="x",
            samplerate=44100,
            channels=2,
            subtype="PCM_24",
        ) as file:
            with sd.InputStream(
                samplerate=44100,
                device="Azure Kinect Microphone , MME",
                channels=2,
                callback=callback,
            ):
                while self.status:
                    capture = azure_device.get_capture()
                    record.write_capture(capture)
                    file.write(q.get())

                    # print(capture.color[:, :, :3])
                    if np.any(capture.color):
                        frame = capture.color[:, :, :3].copy()
                        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

                        h, w, ch = capture.color[:, :, :3].shape
                        img = QImage(frame, w, h, ch * w, QImage.Format_RGB888)
                        scaled_img = img.scaled(640, 480, Qt.KeepAspectRatio)
                        # Emit signal
                        self.updateFrame.emit(scaled_img)

        record.flush()
        record.close()
        print(f"{record.captures_count} frames written.")

        src = self.filename_audio
        dst = f"{self.filename_audio[:-4]}.mp3"

        sound = AudioSegment.from_mp3(src)
        sound.export(dst, format="wav")

        os.remove(self.filename_audio)
        print("finish to save")


class Window(QMainWindow):
    def __init__(self):
        super().__init__()
        # Title and dimensions
        self.setWindowTitle("Patterns detection")
        self.setGeometry(0, 0, 800, 500)

        # Create a label for the display camera
        self.label = QLabel(self)
        self.label.setFixedSize(640, 480)

        # Thread in charge of updating the image
        self.th = Thread(self)
        # self.th.finished.connect(self.close)
        self.th.updateFrame.connect(self.setImage)

        # Buttons layout
        buttons_layout = QHBoxLayout()
        self.button1 = QPushButton("Start")
        self.button2 = QPushButton("Stop/Close")
        self.button1.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        self.button2.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        buttons_layout.addWidget(self.button2)
        buttons_layout.addWidget(self.button1)

        # Main layout
        layout = QVBoxLayout()
        layout.addWidget(self.label)
        layout.addLayout(buttons_layout)
        # layout.addLayout(right_layout)

        # Central widget
        widget = QWidget(self)
        widget.setLayout(layout)
        self.setCentralWidget(widget)

        # Connections
        self.button1.clicked.connect(self.start)
        self.button2.clicked.connect(self.kill_thread)
        self.button2.setEnabled(False)

    @Slot()
    def kill_thread(self):
        print("Finishing...")
        self.button2.setEnabled(False)
        self.button1.setEnabled(True)
        # cv2.destroyAllWindows()
        self.th.status = False
        # Give time for the thread to finish
        time.sleep(1)

    @Slot()
    def start(self):
        print("Starting...")
        self.button2.setEnabled(True)
        self.button1.setEnabled(False)
        self.th.set_filename()
        self.th.status = True
        self.th.start()

    @Slot(QImage)
    def setImage(self, image):
        self.label.setPixmap(QPixmap.fromImage(image))


if __name__ == "__main__":
    q = queue.Queue()
    app = QApplication()
    w = Window()
    w.show()
    sys.exit(app.exec())

    # main(args)



import ctypes

from .k4a import _k4a, Device, default_configuration
from .k4arecord import _k4arecord
from .k4arecord.playback import Playback
from .utils import (
    get_k4a_module_path,
    get_k4abt_module_path,
    get_k4arecord_module_path,
)


def initialize_libraries(
    module_k4a_path=None, module_k4abt_path=None, track_body=False
) -> bool:
    # Search the module path for k4a if not available
    if module_k4a_path is None:
        module_k4a_path = get_k4a_module_path()

    module_k4arecord_path = get_k4arecord_module_path(module_k4a_path)
    _flag = True
    try:
        # Initialize k4a related wrappers
        init_k4a(module_k4a_path)

        # Initialize k4arecord related wrappers
        init_k4arecord(module_k4arecord_path)

        # if track_body:
        #     # Search the module path for k4abt if not available
        #     if module_k4abt_path is None:
        #         module_k4abt_path = get_k4abt_module_path()

        #     # Initialize k4abt related wrappers
        #     init_k4abt(module_k4abt_path)
    except:
        print("Can not setting .dll")
        _flag = False
    
    finally:
        return _flag


def init_k4a(module_k4a_path):
    _k4a.setup_library(module_k4a_path)


# def init_k4abt(module_k4abt_path):
#     _k4abt.setup_library(module_k4abt_path)


def init_k4arecord(module_k4arecord_path):
    _k4arecord.setup_library(module_k4arecord_path)


def start_device(
    device_index=0,
    config=default_configuration,
    record=False,
    record_filepath="output.mkv",
):
    # Create device object
    device = Device(device_index)

    # Start device
    device.start(config, record, record_filepath)
    
    return device


# def start_body_tracker(model_type=_k4abt.K4ABT_DEFAULT_MODEL, calibration=None):
#     if calibration:
#         return Tracker(calibration, model_type)
#     else:
#         return Tracker(Device.calibration, model_type)


def start_playback(filepath):
    return Playback(filepath)



import cv2
import numpy as np


def smooth_depth_image(depth_image, max_hole_size=10):
    """Smoothes depth image by filling the holes using inpainting method

    Parameters:
    depth_image(Image): Original depth image
    max_hole_size(int): Maximum size of hole to fill

    Returns:
    Image: Smoothed depth image

    Remarks:
    Bigger maximum hole size will try to fill bigger holes but requires longer time
    """
    mask = np.zeros(depth_image.shape, dtype=np.uint8)
    mask[depth_image == 0] = 1

    # Do not include in the mask the holes bigger than the maximum hole size
    kernel = np.ones((max_hole_size, max_hole_size), np.uint8)
    erosion = cv2.erode(mask, kernel, iterations=1)
    mask = mask - erosion

    smoothed_depth_image = cv2.inpaint(
        depth_image.astype(np.uint16), mask, max_hole_size, cv2.INPAINT_NS
    )

    return smoothed_depth_image



from .calibration import Calibration
from .capture import Capture
from .configuration import Configuration, default_configuration
from .device import Device
from .image import Image
from .imu_sample import ImuSample
from .transformation import Transformation